// stdafx.h : include file for standard system include files,
// or project specific include files that are used frequently, but
// are changed infrequently
//
#include "stdafx.h"
#include "Resource.h"

#if defined(WCE_VS2005)
#include "..\CEBoard2005\EngineInterfaces.h"
#else
#include "..\EngineInterfaces.h"
#endif

#define NO_MIN_MAX
#define CEBOARD_UCI
#include "CEBoardUCI.h"
#include "posix.h"
#include <time.h>

LPCTSTR ENGINE_KEY = _T("Software\\Alainza\\gfruit4bx");
LPCTSTR g_sEngineName = _T("GambitFruit 1.0 Beta 4bx");
LPCTSTR LOGFILE = _T("\\temp\\gfruit4bx.txt");

DWORD dwEngineStackSize = 1024 * 1024;


// A few declarations (should be cleaned)
extern "C" {
	extern char version[20];
	extern void Print(int, char *, ...);
}

int main(int argc, char *argv[]);

// Will add a \n
static void SendEngineMessage(LPCWSTR message);
// Will not add a \n
static void SendEngineMessage(LPCSTR message);
// Records exchanged messages for debug


static HINSTANCE g_hInstance;


CBuffer *g_pBuffer = NULL;
CUCIEngine *g_pEngine;
BOOL g_bLog = FALSE;
HANDLE g_hLog = INVALID_HANDLE_VALUE;

int errno;

typedef struct tagMoveParam {
	LPCTSTR sLabel;
	LPCTSTR sValue;
	int nValue;
} MoveParam;

static MoveParam g_MoveParams[] = {
	{ _T("All moves"), _T("sd"),	0 },
	{ _T("20 moves"), _T("20"),		20 },
	{ _T("30 moves"), _T("30"),		30 },
	{ _T("40 moves"), _T("40"),		40 },
	{ _T("50 moves"), _T("50"),		50 },
	{ _T("60 moves"), _T("60"),		60 }
};

void WriteLog(char *msg)
{
	if (g_bLog) {
		if (   strncmp(msg, "info currmove", 13)
			&& (strncmp(msg, "info depth", 10) || (strstr(msg, "score") != NULL))
	//info depth 4 seldepth 9 score cp -1445 time 40 nodes 5080 pv g7g6 c5e6 g6h7 c6d5
			) {
			if (g_hLog == INVALID_HANDLE_VALUE) {
				g_hLog = CreateFile(LOGFILE, GENERIC_READ | GENERIC_WRITE, 0, NULL, OPEN_ALWAYS, 0, NULL);
				SetFilePointer(g_hLog, 0, NULL, FILE_END);
			}

			static TCHAR wsbuf[16384];
			MultiByteToWideChar(CP_ACP, 0, msg, -1, wsbuf, 16384);
			_tcscat(wsbuf, _T("\r\n"));

			DWORD dwWritten;
			WriteFile(g_hLog, wsbuf, 2 * _tcslen(wsbuf), &dwWritten, NULL);

			//			CloseHandle(g_hLog);
		}
	}
}

//////////////////////////////////////////////////////////////////////////
//
// "Implementation" of Windows CE CRT Missing functions
//
//////////////////////////////////////////////////////////////////////////
extern "C" {

char *strerror(int err) {
	static char buffer[512];

	switch (err) {
	case 2:		return "No such file or directory";
	case 7:		return "Arg list too long";
	case 8:		return "Exec format error";
	case 9:		return "Bad file number";
	case 12:	return "Not enough core";
	case 13:	return "Permission denied";
	case 17:	return "File exists";
	case 18:	return "Cross-device link";
	case 22:	return "Invalid argument";
	case 24:	return "Too many open files";
	case 28:	return "No space left on device";
	case 33:	return "Math argument";
	case 34:	return "Result too large";
	case 36:	return "Resource deadlock would occur";
	default:
		sprintf(buffer, "Error %d", err);
		return buffer;
	}
}

void __cdecl MyExit(int t) {
	printf("!!!! Sortie de UCIEngineSE %d\n", t);
//	g_pEngine->Exit(t);
	ExitThread(t);
}

void system(char *)
{
	// Just ignore
}

void remove(char *sFichier)
{
	TCHAR wsFichier[MAX_PATH+1];
	MultiByteToWideChar(CP_ACP,0,sFichier,-1,wsFichier,MAX_PATH+1);
	DeleteFile(wsFichier);
}

char *getenv(char *var)
{
	TCHAR wsName[512];
	MultiByteToWideChar(CP_ACP,0,var,-1,wsName,512);
	TCHAR wsValue[512];
	static char sValue[512];
	CRegistryKey engineKey(HKEY_CURRENT_USER, ENGINE_KEY, FALSE);
	if (!engineKey.LireChaine(wsName, wsValue, 512)) {
		return NULL;
	} else {
		WideCharToMultiByte(CP_ACP, 0, wsValue, -1, sValue, 512, NULL, NULL);
		return sValue;
	}
}


void signal(int a, int b)
{
	// No signal here
}

int read(void *canal, char *buffer, int len)
{
	DWORD dwTaille = len;
	if (g_pBuffer->Lire(buffer, &dwTaille, TRUE)) {
		return dwTaille;
	} else {
		return 0;
	}
}

char *fgets(char* string, int n, FILE *stream)
{
	g_pBuffer->Gets(string, n);
	return string;
}


int CheckInput(void)
{
	return g_pBuffer->GetDedans();
}


} // extern C

bool __cdecl input_available(void)
{
	return g_pBuffer->GetDedans() > 0;
}

//////////////////////////////////////////////////////////////////////////
//
// Time related types, variables and functions
//
//////////////////////////////////////////////////////////////////////////
time_t time(time_t *timer);

#if 0 && !defined(WCE_VS2005)
struct tm {
	int tm_sec;	/* seconds after the minute - [0,59] */
	int tm_min;	/* minutes after the hour - [0,59] */
	int tm_hour;	/* hours since midnight - [0,23] */
	int tm_mday;	/* day of the month - [1,31] */
	int tm_mon;	/* months since January - [0,11] */
	int tm_year;	/* years since 1900 */
	int tm_wday;	/* days since Sunday - [0,6] */
	int tm_yday;	/* days since January 1 - [0,365] */
	int tm_isdst;	/* daylight savings time flag */
};
#endif


static time_t g_processAttach;
static __int64 g_januaryFirst1970;
static struct tm g_returnValue;

static void InitTime()
{
	SYSTEMTIME st;
	st.wYear = 1970;
	st.wMonth = 0;
	st.wDay = 1;
	st.wHour = 0;
	st.wMinute = 0;
	st.wSecond = 0;
	st.wMilliseconds = 0;
	SystemTimeToFileTime(&st, (FILETIME*)&g_januaryFirst1970);
	g_processAttach = time(0);
}


extern "C" {

struct tm * __cdecl MyLocaltime(time_t *pSec)
{
	SYSTEMTIME st;
	GetLocalTime(&st);
	g_returnValue.tm_hour = st.wHour;
	g_returnValue.tm_mday = st.wDay;
	g_returnValue.tm_min = st.wMinute;
	g_returnValue.tm_mon = st.wMonth - 1;
	g_returnValue.tm_sec = st.wSecond;
	g_returnValue.tm_wday = st.wDayOfWeek;
	g_returnValue.tm_year = st.wYear - 1900;
	if (pSec) {
		__int64 ft1, diff;
		st.wMilliseconds = 0;
		SystemTimeToFileTime(&st, (FILETIME*)&ft1);
		diff = (ft1 - g_januaryFirst1970) / 1000000000;
		*pSec = (const time_t)diff;
	}
	return &g_returnValue;
}
#define localtime MyLocaltime


time_t time(time_t *timer)
{
	time_t ret;
	localtime(&ret);
	if (timer != NULL) *timer = ret;
	return ret;
}

#if 1 || defined(WCE_VS2005)
clock_t __cdecl clock(void)
#else
int __cdecl clock(void)
#endif
{
	return time(0) - g_processAttach;
}


int    __cdecl printf(const char *sFormat, ...)
{
	if (g_pEngine) {
		static char sBuf[1024];
		int ret;
		va_list argList;
		va_start(argList , sFormat);
		ret = vsprintf(sBuf, sFormat, argList);

		WriteLog(sBuf);
#ifdef DEBUG
		static TCHAR wsBuf[1024];
		MultiByteToWideChar(CP_ACP, 0, sBuf, -1, wsBuf, 1024);
		OutputDebugString(wsBuf);
		OutputDebugString(_T("\n"));
#endif
		g_pEngine->EngineReply(sBuf);
		return ret;
	} else {
		return 0;
	}
}

int vfprintf(FILE *fic,const char *sFormat, va_list argList )
{
	if (g_pEngine) {
		static char sBuf[1024];
		int ret;
		ret = vsprintf(sBuf, sFormat, argList);
		WriteLog(sBuf);
		if (fic == stderr) {
			g_pEngine->FatalError(sBuf);
		} else {
			g_pEngine->EngineReply(sBuf);
		}
		return ret;
	} else {
		return 0;
	}
}

int vprintf( const char *sFormat, va_list argList )
{
	if (g_pEngine) {
		static char sBuf[1024];
		int ret;
		ret = vsprintf(sBuf, sFormat, argList);
		g_pEngine->EngineReply(sBuf);
		return ret;
	} else {
		return 0;
	}
}

} // extern "C"

double now_real        ()
{
   return double(GetTickCount()) / 1000.0;
}

double now_cpu         ()
{
   return double(clock()) / double(CLOCKS_PER_SEC); // OK if CLOCKS_PER_SEC is small enough
}


///////////////////////////////////////////////////////////////////
//
// Communications between controler and engine
//
///////////////////////////////////////////////////////////////////
void EcritLog(char *msg)
{
	TCHAR wsMsg[1024];
	int n = MultiByteToWideChar(CP_ACP, 0, msg, -1, wsMsg, 1024);
	wsMsg[n-1] = _T('\n');
	wsMsg[n] = 0;
	OutputDebugString(wsMsg);
}

static char *g_tabEndOfGame[] = {
	"1-0", "0-1", "1/2-1/2"
};

static BOOL StartsWith(char *line, char *start)
{
	while (*start) {
		if (*line != *start) {
			return FALSE;
		} else {
			line++;
			start++;
		}
	}
	return TRUE;
}

static int CopyTrim(wchar_t *wsBuf, int nDispo, char *src, BOOL *pbCR, BOOL *pbLF, BOOL bTrim)
{
	BOOL bDebut = bTrim;
	BOOL bCR = FALSE;
	BOOL bLF = FALSE;
	int i = 0;
	char c = *src++;
	do {
		if (c == 32) {
			if (!bDebut) {
				wsBuf[i++] = c;
			}
		} else if (c > 32) {
			wsBuf[i++] = c;
			bDebut = FALSE;
		} else if (c == 10) {
			bLF = TRUE;
		} else if (c == 13) {
			bCR = TRUE;
		}
		c = *src++;
	} while ((i < nDispo) && c);
	if (bTrim) {
		// Maintenant, on "trime"  droite...
		while ((i > 0) && (wsBuf[--i] == L' '));
		wsBuf[++i] = 0;
	} else {
		wsBuf[i] = 0;
	}
	*pbCR = bCR;
	*pbLF = bLF;
	return i;
}


#ifdef DEBUG
static void TraceTrim(char *msg)
{
	BOOL bCR, bLF;
	wchar_t wsBuf[512];
	int len = 0;
	wsBuf[len++] = L'<';
	wsBuf[len++] = L'<';
	wsBuf[len++] = L'<';
	len += CopyTrim(wsBuf + len, 512 - len, msg, &bCR, &bLF, TRUE);
	wsBuf[len++] = L'>';
	wsBuf[len++] = L'\n';
	wsBuf[len++] = 0;
	OutputDebugString(wsBuf);
}
#endif
////////////////////////////////////////////////////////////////////////////
CBuffer::CBuffer(DWORD dwTaille)
{
	m_pBuf = new char[dwTaille];
	memset(m_pBuf, ' ', dwTaille);
	m_dwTaille = dwTaille;
	m_PosLecture = 0;
	m_PosEcriture = 0;
	InitializeCriticalSection(&m_cs);
	m_hEventWrite = CreateEvent(NULL, FALSE, FALSE, NULL);
	m_hEventRead = CreateEvent(NULL, FALSE, FALSE, NULL);
}

CBuffer::~CBuffer()
{
	delete [] m_pBuf;
	CloseHandle(m_hEventWrite);
	CloseHandle(m_hEventRead);
	DeleteCriticalSection(&m_cs);
}

DWORD CBuffer::GetDedans(void)
{
	int n = m_PosEcriture - m_PosLecture;
	if (n<0) {
		n += m_dwTaille - 1;
	}
	return n;
}

void CBuffer::Dump(DWORD *pdwPosLecture, DWORD *pdwPosEcriture, char *sBuf)
{
	EnterCriticalSection(&m_cs);
	*pdwPosLecture = m_PosLecture;
	*pdwPosEcriture = m_PosEcriture;
	memcpy(sBuf, m_pBuf, m_dwTaille + 1);
	sBuf[m_dwTaille + 1] = 0;
	LeaveCriticalSection(&m_cs);
}



BOOL CBuffer::Ecrire(const char *sMsg)
{
	{
		OutputDebugString(L">>>");
		TCHAR wsTmp[512];
		MultiByteToWideChar(CP_ACP, 0, sMsg, -1, wsTmp, 512);
		OutputDebugString(wsTmp);
	}
	int nLen, nDispo, nApres;
	EnterCriticalSection(&m_cs);
	nLen = strlen(sMsg);
	nDispo = m_dwTaille - CBuffer::GetDedans();
	if (nLen <= nDispo) {
		// Il faut peut-tre en crire un bout derrire et un bout devant...
		nApres = m_dwTaille - 1 - m_PosEcriture;
		if (nApres >= nLen) {
			// Pas de problme
			memcpy(m_pBuf + m_PosEcriture, sMsg, nLen);
			m_PosEcriture += nLen;
		} else {
			// Il faut en mettre un peu  chaque bout
			memcpy(m_pBuf + m_PosEcriture, sMsg, nApres);
			memcpy(m_pBuf, sMsg + nApres, nLen - nApres);
			m_PosEcriture = nLen - nApres;
		}
		m_PosEcriture = m_PosEcriture % m_dwTaille;
		// On peut partir
		SetEvent(m_hEventRead);	// pour dbloquer les lectures...
		LeaveCriticalSection(&m_cs);
		return TRUE;
	} else {
		// Premier test : on ne fait pas d'criture partielle... pas terrible
		// mais si la taille est bien suprieure aux blocs crits et lus, cela ne
		// doit pas poser de problme
		{
			TCHAR sBuf[512];
			wsprintf(sBuf, _T("Ecriture de %d mais %d disponibles\n"), nLen, nDispo);
			OutputDebugString(sBuf);
		}
		ResetEvent(m_hEventWrite);	// On veut une nouvelle criture
		LeaveCriticalSection(&m_cs);
		WaitForSingleObject(m_hEventWrite, INFINITE);
		return Ecrire(sMsg);
	}
}

void CBuffer::Gets(char *string, int n)
{
	EnterCriticalSection(&m_cs);
	int lus = 0;
	char c = 0;
	while ((lus < n) && (c != '\n')) {
		while (CBuffer::GetDedans() == 0) {
			ResetEvent(m_hEventRead);
			LeaveCriticalSection(&m_cs);
			WaitForSingleObject(m_hEventRead, INFINITE);
			EnterCriticalSection(&m_cs);
		}
		c = m_pBuf[m_PosLecture++];
		if (m_PosLecture == m_dwTaille - 1) {
			m_PosLecture = 0;
		}
		string[lus++] = c;
	}
	string[lus] = 0;
	LeaveCriticalSection(&m_cs);
}

BOOL CBuffer::Lire(char *sMsg, DWORD *pdwTaille, BOOL bWait)
{
	DWORD nApres;
	DWORD nDedans;
	DWORD aLire;

	EnterCriticalSection(&m_cs);
	nDedans = CBuffer::GetDedans();
	if (nDedans == 0) {
		ResetEvent(m_hEventRead);
		LeaveCriticalSection(&m_cs);
		WaitForSingleObject(m_hEventRead, INFINITE);
		EnterCriticalSection(&m_cs);
		nDedans = CBuffer::GetDedans();
	}
	aLire = min(nDedans, *pdwTaille);

	// Peut-tre un peu  chaque bout
	nApres = m_dwTaille - 1 - m_PosLecture;
	if (nApres >= aLire) {
		// Tout va bien
		memcpy(sMsg, m_pBuf + m_PosLecture, aLire);
		m_PosLecture += aLire;
	} else {
		// Il faut le prendre en deux fois
		memcpy(sMsg, m_pBuf + m_PosLecture, nApres);
		memcpy(sMsg + nApres, m_pBuf, aLire - nApres);
		m_PosLecture = aLire - nApres;
	}
	sMsg[aLire] = 0;
	SetEvent(m_hEventWrite);	// Pour dbloquer les critures
	LeaveCriticalSection(&m_cs);
	*pdwTaille = aLire;
	return TRUE;
}

/////////////////////////////////////////////////////////////////

#define MAX_LONG 4096



CRegistryKey::CRegistryKey(HKEY hBranche,LPCTSTR sCle,BOOL bCreer)
{
	LONG nRet;
	DWORD dwRes;

	if (bCreer) {
		nRet = RegCreateKeyEx(hBranche,sCle,0,TEXT("String"),0,
							  KEY_ALL_ACCESS,NULL,&m_hCle,&dwRes);
	} else {
		nRet = RegOpenKeyEx(hBranche,sCle,0,KEY_READ|KEY_ENUMERATE_SUB_KEYS,&m_hCle);
	}
	if (nRet != ERROR_SUCCESS) {
		m_hCle = 0;
	}
}

CRegistryKey::~CRegistryKey()
{
	if (m_hCle != 0)
		RegCloseKey(m_hCle);
}

int CRegistryKey::LireLongueurChaine(LPCTSTR sNom)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		BYTE c;
		DWORD dwType;
		DWORD dwLen = 0;
		if (RegQueryValueEx(m_hCle, sNom, NULL, &dwType, &c, &dwLen) == ERROR_MORE_DATA) {
			return dwLen;
		} else {
			return -1;
		}
	}
}

BOOL CRegistryKey::LireChaine(LPCTSTR sNom, LPTSTR sValeur, DWORD nSize)
{
	DWORD dwType;

	if (m_hCle == 0) {
		return FALSE;
	} else {
		// Lecture de la chane de caractres
		return (ERROR_SUCCESS == RegQueryValueEx(m_hCle,sNom,NULL,&dwType,(LPBYTE)sValeur,&nSize));
	}
}

BOOL CRegistryKey::EcrireChaine(LPCTSTR sNom, LPCTSTR sValeur)
{
	if (m_hCle == 0)
		return FALSE;
	else
		return (ERROR_SUCCESS == RegSetValueEx(m_hCle,
											   sNom,
											   0,
											   REG_SZ,
											   (CONST BYTE*)sValeur,
											   sizeof(TCHAR) * (1 + lstrlen(sValeur))));
}


BOOL CRegistryKey::EnumererChaine(DWORD &nIndice,
								LPTSTR sValeur, DWORD nValSize,
								LPTSTR pName, DWORD nNameSize)
{
	LONG nRet;
	DWORD dwType;

	if (m_hCle == 0) {
		return FALSE;
	} else {
		// Lecture de la chane de caractres
		nRet = RegEnumValue(m_hCle,nIndice,pName,&nNameSize,NULL,
							&dwType,(LPBYTE)sValeur,&nValSize);
		if (nRet == ERROR_SUCCESS) {
			nIndice++;
			return TRUE;
		} else {
			// Rcupration de la valeur impossible, mais on ne va pas
			// ennuyer l'utilisateur pour cela
			return FALSE;
		}
	}
}

BOOL CRegistryKey::VerifierExistence(LPCTSTR sVal)
{
	LONG nRet;
	DWORD dwType;
	TCHAR sTmp[8];
	DWORD nTaille = sizeof(sTmp);

	if (sVal == NULL) {
		// Vrification sur la cl elle-mme
		return (m_hCle != 0);
	} else {
		// Recherche d'une sous-cl
		if (m_hCle == 0) {
			return FALSE;
		} else {
			// Lecture de la chane de caractres
			nRet = RegQueryValueEx(m_hCle,sVal,NULL,&dwType,(LPBYTE)sTmp,&nTaille);
			// La cl existe si la lecture s'est bien passe ou si elle a chou
			// pour un problme de taille des donnes
			return (nRet == ERROR_SUCCESS) || (nRet == ERROR_MORE_DATA);
		}
	}
}


BOOL CRegistryKey::LireValeurDefaut(LPTSTR sValeur, DWORD nSize)
{
	DWORD dwType;
	if (m_hCle == 0) {
		return FALSE;
	} else {
		return (ERROR_SUCCESS == RegQueryValueEx(m_hCle,
												 NULL,
												 NULL,
												 &dwType,
												 (LPBYTE)sValeur,
												 &nSize));
	}
}

BOOL CRegistryKey::EcrireValeurDefaut(LPCTSTR sValeur)
{
	return EcrireChaine(TEXT(""), sValeur);
}


BOOL CRegistryKey::EnumererSousCles(DWORD &nIndice, LPTSTR pName, DWORD nNameSize)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		static TCHAR sClasse[MAX_LONG];
		DWORD dwTailleClasse = MAX_LONG;
		FILETIME ti;

		return (ERROR_SUCCESS == RegEnumKeyEx(m_hCle,
											  nIndice++,
											  pName,
											  &nNameSize,
											  NULL,
											  sClasse,
											  &dwTailleClasse,
											  &ti));

	}
}

BOOL CRegistryKey::LireLong(LPCTSTR sNom, long *pnVal)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		DWORD dwType;
		DWORD nTaille = sizeof(long);
		return (ERROR_SUCCESS == RegQueryValueEx(m_hCle,sNom,NULL,&dwType,(unsigned char *)pnVal,&nTaille));
	}		
}

BOOL CRegistryKey::EcrireLong(LPCTSTR sNom, long nVal)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		return (ERROR_SUCCESS == RegSetValueEx(m_hCle, sNom, 0, REG_DWORD, (BYTE*)&nVal, sizeof(nVal)));
	}
}

BOOL CRegistryKey::LireFloat(LPCTSTR sNom, float *pfVal)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		DWORD dwType;
		DWORD nTaille = sizeof(float);
		return (ERROR_SUCCESS == RegQueryValueEx(m_hCle,sNom,NULL,&dwType,(unsigned char *)pfVal,&nTaille));
	}		
}

BOOL CRegistryKey::EcrireFloat(LPCTSTR sNom, float fVal)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		return (ERROR_SUCCESS == RegSetValueEx(m_hCle, sNom, 0, REG_DWORD, (BYTE*)&fVal, sizeof(fVal)));
	}
}


BOOL CRegistryKey::LireBlob(LPCTSTR sNom, LPVOID lpBlob, DWORD dwSize, DWORD *pdwRead)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		DWORD dwType;
		LONG nRet = RegQueryValueEx(m_hCle, sNom, 0, &dwType, (LPBYTE)lpBlob, &dwSize);
		if (pdwRead)
			*pdwRead = dwSize;
		return (nRet == ERROR_SUCCESS);
	}
}


BOOL CRegistryKey::EcrireBlob(LPCTSTR sNom, LPVOID lpBlob, DWORD dwSize)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		return (ERROR_SUCCESS == RegSetValueEx(m_hCle, sNom, 0, REG_BINARY, (LPBYTE)lpBlob, dwSize));
	}
}


BOOL CRegistryKey::Delete(HKEY hBranche, LPCTSTR sCle)
{
	// normalement, on devrait tester l'OS : sous 95, la suppression
	// est automatiquement rcursive : on devrait alors dlguer la rcursion
	// pour ne pas faire le test  chaque niveau

	TCHAR sName[MAX_PATH];
	DWORD dwTaille;
	DWORD dwIndex = 0;
	FILETIME ftime;
	HKEY hTmp;

	// on va itrer pour supprimer les sous-cls
	if (RegOpenKeyEx(hBranche, sCle, 0, KEY_ALL_ACCESS, &hTmp) != ERROR_SUCCESS)
		return FALSE;

	while (1) {
		dwTaille = MAX_PATH;
		if (ERROR_NO_MORE_ITEMS == RegEnumKeyEx(hTmp, dwIndex, sName, &dwTaille, 0, NULL, NULL, &ftime)) {
			break;
		}
		if (!Delete(hTmp, sName))
			return FALSE;
	}
	RegCloseKey(hTmp);

	return RegDeleteKey(hBranche, sCle) == ERROR_SUCCESS;
}


// Supprime une valeur
BOOL CRegistryKey::SupprimerValeur(LPCTSTR sNom)
{
	if (m_hCle == 0) {
		return FALSE;
	} else {
		return (ERROR_SUCCESS == RegDeleteValue(m_hCle, sNom));
	}
}




///////////////////////////////////////////////////////////////////////////////////////////////////

BOOL WINAPI DllMain(HANDLE hinstDLL, DWORD dwReason, LPVOID lpvReserved)
{
	switch (dwReason) {
	case DLL_PROCESS_ATTACH:
		g_hInstance = (HINSTANCE)hinstDLL;
		InitTime();
		g_pBuffer = new CBuffer(8192);
		break;
	case DLL_PROCESS_DETACH:
		delete g_pBuffer;
		break;
	}
	return TRUE;
}


DWORD WINAPI EngineControlProc(LPVOID lpParameter)
{
	main(0, NULL);
	return 0;
}


extern "C" IEngine __declspec(dllexport) * LoadEngine(ICEBoard20 *pCEBoard)
{
#if 0
//	char *ligne = "position startpos moves e2e4 d7d5 e4d5 g8f6 d2d4 f6d5 c2c4 d5f6 b1c3 b8c6 g1f3 c8g4 f1e2 e7e6 e1g1 f8b4 c1g5 h7h6 g5h4 e8g8 d4d5 e6d5 c3d5 b4e7 h4g3 f6d5 c4d5 c6b4 e2c4 e7d6 a2a3 g4f3 d1f3 b4c2 a1d1 d8g5 g3d6 c7d6 f3e2 a8c8 e2c2 b7b5 b2b3 b5c4 b3c4 g5f4 d1c1 f8e8 c2d3 e8e4 g2g3 f4e5 d3b3 a7a5 b3b6 e4c4 c1c4 c8c4 b6a5 c4c5 a5a8 g8h7 a8b7 e5d5 b7d5 c5d5 f1b1 d5a5 b1b3 h7g6 f2f4\n";
	char *ligne = "abcdefghijklmnopqrstuvwxyz\n";
	int longueur1 = strlen(ligne);
	char lecture[512];
	g_pBuffer->Ecrire(ligne);
	fgets(lecture, 512, NULL);
	int longueur2 = strlen(lecture);
	g_pBuffer->Ecrire(ligne);
	fgets(lecture, 512, NULL);
	longueur2 = strlen(ligne);

	return NULL;
#else
	srand(GetTickCount());
	return new CUCIEngine(pCEBoard);
#endif
}


CUCIEngine::~CUCIEngine()
{
	CloseHandle(m_hEngineThread);
	CloseHandle(m_hReplyReceived);
}

CUCIEngine::CUCIEngine(ICEBoard20 *pCEBoard)
{
	m_hReplyReceived = CreateEvent(NULL, FALSE, FALSE, NULL);
	m_pCEBoard = pCEBoard;
	g_pEngine = this;
	DWORD dwThreadID;
	m_hEngineThread = CreateThread(NULL, dwEngineStackSize, EngineControlProc, NULL, STACK_SIZE_PARAM_IS_A_RESERVATION, &dwThreadID);
	WaitFor("uci\n", "uciok");
	// Set Hashtable size to 4 MB
	m_nGamePointer = 0;
	m_bWantMessages = FALSE;
	m_hMenu = NULL;
	m_bAnalyze = FALSE;
	m_nDepth = 0;

	// Initialize Engine Parameters
	CRegistryKey cle(HKEY_CURRENT_USER, ENGINE_KEY, FALSE);
	// Time options
	if (!cle.LireLong(L"TimeControlType", &m_nTimeControlType)) m_nTimeControlType = 0;
	if (!cle.LireLong(L"GameTime", &m_nGameTime)) m_nGameTime = 25;
	if (!cle.LireLong(L"MoveCount", &m_nMoveCount)) m_nMoveCount = 0;
	if (!cle.LireLong(L"MoveIncrement", (long*)&m_bMoveIncrement)) m_bMoveIncrement = FALSE;
	if (!cle.LireLong(L"IncrementValue", &m_nIncrementValue)) m_nIncrementValue = 10;
	if (!cle.LireLong(L"FixedTime", &m_nFixedTime)) m_nFixedTime = 4;
	// Other options
	if (!cle.LireLong(L"Hash", &m_nHash)) m_nHash = 4;
	if (!cle.LireChaine(L"Book", m_sBookPath, MAX_PATH)) {
		int index = GetModuleFileName(g_hInstance, m_sBookPath, MAX_PATH);
		while ((index > 0) && (m_sBookPath[index-1] != '\\')) {
			index--;
		}
		_tcscpy(m_sBookPath + index, _T("book_small.bin"));
	}
	if (!cle.LireLong(L"Log", (long*)&g_bLog)) g_bLog = FALSE;

	SendOtherOptionsCommands(TRUE, m_nHash, m_sBookPath, g_bLog);

}

BOOL CUCIEngine::GetProperties(EngineProperties *pProps)
{
	if (pProps->dwSize < sizeof(EngineProperties)) {
		return FALSE;
	}
	pProps->dwVersion = ENGINE_INTERFACE_VERSION;
	pProps->dwCaps = MOVES_UCI;
	_tcsncpy(pProps->sEngineName, g_sEngineName, sizeof(pProps->sEngineName) / sizeof(pProps->sEngineName[0]));
	return TRUE;
}

void CUCIEngine::Release()
{
	EndGame();
	SendEngineMessage("quit\n");
	HCURSOR hOld = SetCursor(LoadCursor(NULL, IDC_WAIT));
	WaitForSingleObject(m_hEngineThread, 2000);
	SetCursor(hOld);
	g_pEngine = NULL;
	DestroyMenu(m_hMenu);
	m_hMenu = NULL;
	if (g_hLog != INVALID_HANDLE_VALUE) {
		CloseHandle(g_hLog);
		g_hLog = INVALID_HANDLE_VALUE;
	}
	delete this;
}

void CUCIEngine::StartGame(BOOL bPlayBlack, BOOL bAnalyze, LPCTSTR sFen)
{
	m_bFinished = FALSE;
	SetThinking(FALSE);
	SendEngineMessage("newgameuci\n");
	WaitFor("isready\n", "readyok");


	if (sFen != NULL) {
		SetPosition(sFen);
	} else {
		m_bAddMoves = TRUE;
		strcpy(m_sGame, "position startpos\n");
		m_nGamePointer = strlen(m_sGame) - 1; // Ignore \n
	}

	m_bAnalyze = bAnalyze;
	if (bAnalyze) {
		SendEngineMessage(m_sGame);
		SendEngineMessage("go infinite\n");
	} else {
		if (bPlayBlack) {
			// Nothing special to do as the user will send his move later
		} else {
			MoveNow();
		}
	}
}

void CUCIEngine::AddMove(LPCTSTR sMove)
{
	if (m_bAddMoves) {
		strcpy(m_sGame + m_nGamePointer, " moves");
		m_nGamePointer += 6;
		m_bAddMoves = FALSE;
	}
	m_sGame[m_nGamePointer++] = ' ';
	m_nGamePointer += WideCharToMultiByte(CP_ACP,0, sMove, -1, m_sGame + m_nGamePointer, GAME_LEN - m_nGamePointer, NULL, NULL) - 1;
	strcpy(m_sGame + m_nGamePointer, "\n");
}

void CUCIEngine::SendMove(LPCTSTR sMove)
{
	if (m_bAnalyze) {
		SendEngineMessage("stop\n");
	}

	// Add move to position buffer
	AddMove(sMove);
	SendEngineMessage(m_sGame);

	if (m_bAnalyze) {
		SendEngineMessage("go infinite\n");
	} else {
		TCHAR sCommand[128];
		// Calcul du temps de jeu
		if (m_nTimeControlType == 0) {
			// On interroge la pendule
			int wtime, btime, movestogo;
			m_pCEBoard->GetTimeInfo(&wtime, &btime, &movestogo);
			int inc = 0;
			if (m_bMoveIncrement) {
				inc = m_nIncrementValue;
			}
			wsprintf(sCommand, _T("go wtime %d btime %d winc %d binc %d movestogo %d"), wtime, btime, inc, inc, movestogo);
		} else {
			wsprintf(sCommand, _T("go movetime %d"), m_nFixedTime * 1000);
		}
		SendEngineMessage(sCommand);
	}
	SetThinking(TRUE);
}


void CUCIEngine::MoveNow()
{
	if (m_bThinking) {
		// TODO : handle switch side if engine is not thinking...
		SendEngineMessage("stop\n");
	} else {
		SetThinking(TRUE);
		// C'est un changement de ct
		if (m_pCEBoard) m_pCEBoard->SwitchSides();

		// Send the current position to the engine
		SendEngineMessage(m_sGame);

		TCHAR sCommand[128];
		// Calcul du temps de jeu
		if (m_nTimeControlType == 0) {
			// On interroge la pendule
			int wtime, btime, movestogo;
			int inc = 0;
			if (m_bMoveIncrement) {
				inc = m_nIncrementValue;
			}
			m_pCEBoard->GetTimeInfo(&wtime, &btime, &movestogo);
			wsprintf(sCommand, _T("go wtime %d btime %d winc %d binc %d movestogo %d"), wtime, btime, inc, inc, movestogo);
		} else {
			wsprintf(sCommand, _T("go movetime %d"), m_nFixedTime * 1000);
		}
		SendEngineMessage(sCommand);
	}
}

void CUCIEngine::EndGame()
{
	m_bFinished = TRUE;
	if (m_bThinking || m_bAnalyze) {
		SendEngineMessage("stop\n");
	}
}

void FatalError(const char *msg)
{
	if (g_pEngine) {
		g_pEngine->FatalError(msg);
	}
}

void CUCIEngine::FatalError(const char *msg)
{
	if (!m_bFinished && m_pCEBoard) {
		TCHAR wsMsg[512];
		MultiByteToWideChar(CP_ACP,0,msg,-1,wsMsg,512);
		m_pCEBoard->FatalError(wsMsg);
	}
}

void CUCIEngine::WaitFor(LPCSTR sCommand, LPCSTR sReply)
{
	strcpy(m_sReply, sReply);
	SendEngineMessage(sCommand);
	InterlockedExchange((LPLONG)&m_nReplyLength, strlen(sReply));
	if (WaitForSingleObject(m_hReplyReceived, 10000) != WAIT_OBJECT_0) {
		m_pCEBoard->FatalError(_T("Engine did not answer in expected time"));
	}
	InterlockedExchange((LPLONG)&m_nReplyLength, 0);
}

void CUCIEngine::EngineReply(char *msg)
{
	// 32 bits reading should be atomic (need to check that)
	if (m_nReplyLength) {
		// Is it the command we are waiting for ?
		if (!strncmp(msg, m_sReply, m_nReplyLength)) {
			// Yes
			SetEvent(m_hReplyReceived);
		}
	}
	{
		TCHAR sBuf[512];
		MultiByteToWideChar(CP_ACP, 0, msg, -1, sBuf, 512);
		OutputDebugString(sBuf);
	}
	if (m_pCEBoard) {
		if ((!m_bAnalyze) && StartsWith(msg, "bestmove")) {
			char *moveStart = msg + 9;
			char *moveEnd = strchr(moveStart, ' ');
			int nb;
			if (moveEnd) {
				nb = (moveEnd - moveStart);
			} else {
				nb = strlen(moveStart) - 1; // Ignore the trailing '\n'
			}

			TCHAR sBuf[128];
			MultiByteToWideChar(CP_ACP, 0, moveStart, nb, sBuf, 128);
			sBuf[nb] = 0;
			AddMove(sBuf);
			m_pCEBoard->EngineMove(sBuf);
			// Renvoie immdiatement la position au moteur pour dtecter les fins de partie
			SendEngineMessage(m_sGame);
			SetThinking(FALSE);
		} else if (m_bWantMessages) {
			if (StartsWith(msg, "info time")) {
				// info time 1050 nodes 70000 nps 66667 cpuload 0
				int nTime = atol(msg + 10);
				char *pStart = strstr(msg + 10, "nodes");
				if (!pStart) return;
				int totalNodes = atol(pStart + 6);
				pStart = strstr(pStart + 6, "nps");
				if (!pStart) return;
				int nodesSec = atol(pStart + 4);
				m_pCEBoard->EngineFeedback(m_nDepth, nodesSec / 1000, totalNodes, nTime, NULL, NULL);
			} else if (StartsWith(msg, "info multipv 1 ")) {
				// info depth 42 seldepth 5 score mate -2 time 8860 nodes 3147554 pv g6h7 f4g5 h7h8 g5g7
				// info depth 4 seldepth 9 score cp -1445 time 40 nodes 5080 pv g7g6 c5e6 g6h7 c6d5
				// Tentative de parsing de la chane...
				char *pStart = strstr(msg, "score");
				if (pStart) {
					// this is an interesting info
					int nodesSec = 0;
					int totalNodes = 0;
					int nTime = 0;
					char *pEval = NULL;
					char *pLine = NULL;
					pEval = pStart + 6;
					char *pEnd = strstr(pEval, "time");
					if (!pEnd) return;
					*pEnd = 0;
					pStart = strstr(msg, "depth");
					if (!pStart) return;
					m_nDepth = atol(pStart + 6);
					// Now go to the right
					nTime = atol(pEnd + 4);
					pStart = strstr(pEnd + 5, "nodes");
					if (!pStart) return;
					totalNodes = atol(pStart + 6);
					if (nTime > 0) {
						nodesSec = totalNodes / nTime;
					}
					pStart = strstr(pStart + 6, "pv");
					pLine = pStart + 3;
					pEnd = strchr(pLine, '\n');
					if (pEnd) {
						*pEnd = 0;
					}
					m_pCEBoard->EngineFeedback(m_nDepth, nodesSec, totalNodes, nTime, pEval, pLine);
				}
			}
		}
	}
}


// Will truncate the message to 512 but not crash
void SendEngineMessage(LPCTSTR message)
{
	char tmp[512];
	int ret = WideCharToMultiByte(CP_ACP, 0, message, -1, tmp, 511, NULL, NULL);
	tmp[ret-1] = '\n';
	tmp[ret] = 0;
	g_pBuffer->Ecrire(tmp);
}

void SendEngineMessage(LPCSTR message)
{
	g_pBuffer->Ecrire(message);
}


void CUCIEngine::CheckShowAnalysis()
{
	CheckMenuItem(m_hMenu, ID_ENGINE_SHOWANALYSIS, MF_BYCOMMAND | (m_bWantMessages ? MF_CHECKED : MF_UNCHECKED));
}

HMENU CUCIEngine::GetMenu()
{
	if (!m_hMenu) {
		m_hMenu = LoadMenu(g_hInstance, MAKEINTRESOURCE(IDR_MENUBAR1));
	}
	// Il faut maintenant vrifier la compatibilit de certaines commandes et du mode slectionn
	EnableMenuItem(m_hMenu, ID_ENGINE_MOVENOW, MF_BYCOMMAND | (m_bAnalyze ? MF_GRAYED : MF_ENABLED));
	EnableMenuItem(m_hMenu, ID_ENGINE_SHOWANALYSIS, MF_BYCOMMAND | (m_bAnalyze ? MF_GRAYED : MF_ENABLED));
	SetThinking(m_bThinking);
	CheckShowAnalysis();
	return GetSubMenu(m_hMenu, 0);
}

static BOOL CALLBACK AboutProc(HWND hDlg, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
	switch (nMsg) {
	case WM_CLOSE:
	case WM_COMMAND:
		EndDialog(hDlg, IDOK);
		break;
	case WM_INITDIALOG:
		{
			SHINITDLGINFO shidi;
			// Create a Done button and size it.
			shidi.dwMask = SHIDIM_FLAGS;
			shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIPDOWN | SHIDIF_SIZEDLGFULLSCREEN | SHIDIF_EMPTYMENU;
			shidi.hDlg = hDlg;
			//initialzes the dialog based on the dwFlags parameter
			SHInitDialog(&shidi);
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

BOOL CUCIEngine::HandleCommand(HWND hWnd, WPARAM wParam, LPARAM lParam)
{
	switch (LOWORD(wParam)) {
	case ID_MENU_ENGINE_ABOUT:
		DialogBoxParam(g_hInstance, MAKEINTRESOURCE(IDD_ABOUTBOX), hWnd, AboutProc, NULL);
		break;
	case ID_MENU_ENGINE_OPTIONS:
		ShowOptions(hWnd);
		break;
	case ID_ENGINE_MOVENOW:
		MoveNow();
		break;
	case ID_ENGINE_SHOWANALYSIS:
		m_bWantMessages = !m_bWantMessages;
		CheckShowAnalysis();
		if (m_pCEBoard) m_pCEBoard->ShowThinking(m_bWantMessages);
		break;
	default:
		return FALSE;
	}
	return TRUE;
}

void CUCIEngine::SetPosition(LPCTSTR sFEN)
{
	if (m_bAnalyze) {
		SendEngineMessage("stop\n");
	}
	m_bAddMoves = TRUE;
	strcpy(m_sGame, "position fen ");
	m_nGamePointer = strlen(m_sGame);
	m_nGamePointer += WideCharToMultiByte(CP_ACP, 0, sFEN, -1, m_sGame + m_nGamePointer, GAME_LEN - m_nGamePointer, NULL, NULL) - 1;
	strcpy(m_sGame + m_nGamePointer, "\n");

	if (m_bAnalyze) {
		SendEngineMessage(m_sGame);
		SendEngineMessage("go infinite\n");
	}
}

BOOL CUCIEngine::IsAlive()
{
	DWORD dwRet = WaitForSingleObject(m_hEngineThread, 0);
	return dwRet != WAIT_OBJECT_0;
}

void CUCIEngine::GetMessages(BOOL bWantMessages)
{
	m_bWantMessages = bWantMessages;
}


void CUCIEngine::SendTimeCommands(BOOL bForce, long nTimeControlType, long nGameTime, long nMoveCount,
								  BOOL bMoveIncrement, long nIncrementValue, long nFixedTime)
{
	if (bForce || (nTimeControlType != m_nTimeControlType)) {
		m_nTimeControlType = nTimeControlType;
	}
	if (bForce || (nGameTime != m_nGameTime)) {
		m_nGameTime = nGameTime;
	}
	if (bForce || (nMoveCount != m_nMoveCount)) {
		m_nMoveCount = nMoveCount;
	}
	if (bForce || (bMoveIncrement != m_bMoveIncrement)) {
		m_bMoveIncrement = bMoveIncrement;
	}
	if (bForce || (nIncrementValue != m_nIncrementValue)) {
		m_nIncrementValue = nIncrementValue;
	}
	if (bForce || (nFixedTime != m_nFixedTime)) {
		m_nFixedTime = nFixedTime;
	}

	// Now tell to the host
	if (nTimeControlType == 0) {
		// Normal time control
		m_pCEBoard->SetTimeControl(0, nMoveCount, nGameTime * 60, 0, 0, 0, m_bMoveIncrement ? nIncrementValue : 0);
	} else {
		m_pCEBoard->SetTimeControl(nFixedTime, 0, 0, 0, 0, 0, 0);
	}
}

void CUCIEngine::SendOtherOptionsCommands(BOOL bForce, int nHash, LPCTSTR sBook, BOOL bLog)
{
	TCHAR sCommand[MAX_PATH + 32];
	if (bForce || (nHash != m_nHash)) {
		m_nHash = nHash;
		wsprintf(sCommand, _T("setoption name Hash value %d"), m_nHash);
		SendEngineMessage(sCommand);
	}
	if (bForce || _tcscmp(m_sBookPath, sBook)) {
		_tcscpy(m_sBookPath, sBook);
		if (m_sBookPath[0]) {
			_tcscpy(sCommand, _T("setoption name BookFile value "));
			_tcscat(sCommand, m_sBookPath);
			SendEngineMessage(sCommand);
			SendEngineMessage("setoption name OwnBook value true\n");
		} else {
			SendEngineMessage("setoption name OwnBook value false\n");
		}
	}
	if (bForce || (bLog != g_bLog)) {
		g_bLog = bLog;
		if (bLog) {
			BYTE bom[2] = { 0xFE, 0xFF };
			HANDLE hFic = CreateFile(LOGFILE, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, 0, NULL);
			DWORD dwWritten;
			WriteFile(hFic, bom, 2, &dwWritten, NULL);
			CloseHandle(hFic);
		}
	}
	// Dsactivation multipv
//	SendEngineMessage("setoption name MultiPV value 0\n");
}


///////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//
// User interface
static void InvalidOption(HWND hDlg, LPCTSTR sMessage)
{
	MessageBox(hDlg, sMessage, _T("Invalid parameter"), MB_OK | MB_ICONERROR);
}

static int CALLBACK MaPropSheetProc(HWND hDlg, UINT uMsg, LPARAM lParam)
{
	switch (uMsg) {
	case PSCB_INITIALIZED:
		{
			SHINITDLGINFO shidi;
			// Create a Done button and size it.
			shidi.dwMask = SHIDIM_FLAGS;
			shidi.dwFlags = SHIDIF_DONEBUTTON | SHIDIF_SIZEDLG | SHIDIF_EMPTYMENU;
			shidi.hDlg = hDlg;
			//initialzes the dialog based on the dwFlags parameter
			SHInitDialog(&shidi);
		}
		break;
	case PSCB_PRECREATE:
		break; // rien a priori
	}
	return 0;
}


// Ouverture de la bote de dialogue
BOOL CUCIEngine::ShowOptions(HWND hParent)
{
	PROPSHEETPAGE pages[2];
	ZeroMemory(pages, sizeof(pages));
	pages[0].dwSize = sizeof(PROPSHEETPAGE);
	pages[0].dwFlags = PSP_DEFAULT;
	pages[0].hInstance = g_hInstance;
	pages[0].pszTemplate = MAKEINTRESOURCE(IDD_TIMEOPTIONS);
	pages[0].pfnDlgProc = GlobalTimeControlProc;
	pages[0].lParam = (LPARAM)this;

	pages[1].dwSize = sizeof(PROPSHEETPAGE);
	pages[1].dwFlags = PSP_DEFAULT;
	pages[1].hInstance = g_hInstance;
	pages[1].pszTemplate = MAKEINTRESOURCE(IDD_MISCOPTIONS);
	pages[1].pfnDlgProc = GlobalOtherOptionsProc;
	pages[1].lParam = (LPARAM)this;


	PROPSHEETHEADER psh;
	ZeroMemory(&psh, sizeof(psh));
	psh.dwSize = sizeof(psh);
	psh.dwFlags = PSH_PROPSHEETPAGE | PSH_USECALLBACK;
	psh.pfnCallback = MaPropSheetProc;
	psh.hwndParent = hParent;
	psh.hInstance = g_hInstance;
	psh.pszCaption = _T("Engine options");
	psh.nStartPage = 0;
	psh.ppsp = pages;
	psh.nPages = 2;
	int nRet = PropertySheet(&psh);
	if (nRet == IDOK) {
		// Options have been sent and saved in members
		// Save them in registry
		CRegistryKey cle(HKEY_CURRENT_USER, ENGINE_KEY, TRUE);
		// Time options
		cle.EcrireLong(L"TimeControlType", m_nTimeControlType);
		cle.EcrireLong(L"GameTime", m_nGameTime);
		cle.EcrireLong(L"MoveCount", m_nMoveCount);
		cle.EcrireLong(L"MoveIncrement", m_bMoveIncrement);
		cle.EcrireLong(L"IncrementValue", m_nIncrementValue);
		cle.EcrireLong(L"FixedTime", m_nFixedTime);
		// Other options
		cle.EcrireLong(L"Hash", m_nHash);
		cle.EcrireChaine(L"Book", m_sBookPath);
		// Now tell the host the choosen time control

		return TRUE;
	} else {
		return FALSE;
	}
}

// Procdure de rappel Windows
BOOL CALLBACK CUCIEngine::GlobalTimeControlProc(HWND hDlg, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
	CUCIEngine *pThis;
	if (nMsg == WM_INITDIALOG) {
		PROPSHEETPAGE *pPSP = (PROPSHEETPAGE *)lParam;
		pThis = (CUCIEngine*)pPSP->lParam;
		SetWindowLong(hDlg, DWL_USER, pPSP->lParam);
	} else {
		pThis = (CUCIEngine*)GetWindowLong(hDlg, DWL_USER);
	}
	if (pThis) {
		return pThis->TimeControlProc(hDlg, nMsg, wParam, lParam);
	} else {
		return FALSE;
	}
}

void CUCIEngine::SetTimeValues(HWND hDlg)
{
	CheckDlgButton(hDlg, IDC_RB_TIMECONTROL, (m_nTimeControlType == 0) ? BST_CHECKED : BST_UNCHECKED);
	CheckDlgButton(hDlg, IDC_RB_ABSOLUTETIME, (m_nTimeControlType == 1) ? BST_CHECKED : BST_UNCHECKED);

	SendDlgItemMessage(hDlg, IDC_CBX_MOVECOUNT, CB_SETCURSEL, m_nMoveCount, 0);
	SetDlgItemInt(hDlg, IDC_EDT_GAMETIME, m_nGameTime, TRUE);
	CheckDlgButton(hDlg, IDC_CHK_INCREMENT, m_bMoveIncrement);
	SetDlgItemInt(hDlg, IDC_EDT_INCREMENTVALUE, m_nIncrementValue, TRUE);

	SetDlgItemInt(hDlg, IDC_EDT_TIMEMOVE, m_nFixedTime, TRUE);
}

BOOL CUCIEngine::ReadTimeValues(HWND hDlg, long *pnTimeControlType, long *pnGameTime, long *pnMoveCount, BOOL *pbMoveIncrement, long *pnIncrementValue, long *pnFixedTime)
{
	*pnTimeControlType = (IsDlgButtonChecked(hDlg, IDC_RB_TIMECONTROL) == BST_CHECKED) ? 0 : 1;
	*pnGameTime = GetDlgItemInt(hDlg, IDC_EDT_GAMETIME, NULL, TRUE);
	int indice = SendDlgItemMessage(hDlg, IDC_CBX_MOVECOUNT, CB_GETCURSEL, 0, 0);
	if ((indice >= 0) && (indice < (sizeof(g_MoveParams) / sizeof(g_MoveParams[0])))) {
		*pnMoveCount = g_MoveParams[indice].nValue;
	} else {
		*pnMoveCount = 0;
	}
	*pbMoveIncrement = IsDlgButtonChecked(hDlg, IDC_CHK_INCREMENT);
	*pnIncrementValue = GetDlgItemInt(hDlg, IDC_EDT_INCREMENTVALUE, NULL, TRUE);
	*pnFixedTime = GetDlgItemInt(hDlg, IDC_EDT_TIMEMOVE, NULL, TRUE);

	return TRUE;
}

BOOL CUCIEngine::TimeControlProc(HWND hDlg, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
	int i;
	switch (nMsg) {
	case WM_INITDIALOG:
		{
			// Prepare the controls
			HWND hCbxMove1 = GetDlgItem(hDlg, IDC_CBX_MOVECOUNT);
			for (i=0;i<sizeof(g_MoveParams)/sizeof(g_MoveParams[0]);i++) {
				SendMessage(hCbxMove1, CB_ADDSTRING, 0, (LPARAM)g_MoveParams[i].sLabel);
			}
			SendDlgItemMessage(hDlg, IDC_SPIN_GAMETIME, UDM_SETRANGE, 0, MAKELONG(180, 1));
			SendDlgItemMessage(hDlg, IDC_SPIN_INCREMENT, UDM_SETRANGE, 0, MAKELONG(180, 0));
			SendDlgItemMessage(hDlg, IDC_SPIN_FIXED,UDM_SETRANGE, 0, MAKELONG(1800, 1));

			// Now fill the controls
			SetTimeValues(hDlg);

		}
		break;
	case WM_DESTROY:
		{
			// Read controls values
			long nTimeControlType, nGameTime, nMoveCount, nIncrementValue, nFixedTime;
			BOOL bMoveIncrement;

			ReadTimeValues(hDlg, &nTimeControlType, &nGameTime, &nMoveCount, &bMoveIncrement, &nIncrementValue, &nFixedTime);

			SendTimeCommands(FALSE, nTimeControlType, nGameTime, nMoveCount, bMoveIncrement, nIncrementValue, nFixedTime);

		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_RB_TIMECONTROL:
		case IDC_RB_ABSOLUTETIME:
			{
				BOOL bNormalTimeControl = IsDlgButtonChecked(hDlg, IDC_RB_TIMECONTROL);
				EnableWindow(GetDlgItem(hDlg, IDC_CBX_MOVECOUNT), bNormalTimeControl);
				EnableWindow(GetDlgItem(hDlg, IDC_EDT_GAMETIME), bNormalTimeControl);
				EnableWindow(GetDlgItem(hDlg, IDC_SPIN_GAMETIME), bNormalTimeControl);
				EnableWindow(GetDlgItem(hDlg, IDC_CHK_INCREMENT), bNormalTimeControl);
				EnableWindow(GetDlgItem(hDlg, IDC_EDT_INCREMENTVALUE), bNormalTimeControl);
				EnableWindow(GetDlgItem(hDlg, IDC_SPIN_INCREMENT), bNormalTimeControl);
				EnableWindow(GetDlgItem(hDlg, IDC_SPIN_FIXED), !bNormalTimeControl);
				if (bNormalTimeControl) {
					TimeControlProc(hDlg, WM_COMMAND, IDC_CHK_INCREMENT, 0);
					TimeControlProc(hDlg, WM_COMMAND, IDC_CBX_MOVECOUNT, 0);
				}

			}
			break;
		case IDC_CHK_INCREMENT:
			{
				BOOL bIncrementEnabled = (IsDlgButtonChecked(hDlg, IDC_CHK_INCREMENT));
				EnableWindow(GetDlgItem(hDlg, IDC_EDT_INCREMENTVALUE), bIncrementEnabled);
				EnableWindow(GetDlgItem(hDlg, IDC_SPIN_INCREMENT), bIncrementEnabled);
			}
			break;
		default:
			return FALSE;
		}
		break;
	default:
		return FALSE;
	}
	return TRUE;
}


BOOL CALLBACK CUCIEngine::GlobalOtherOptionsProc(HWND hDlg, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
	CUCIEngine *pThis;
	if (nMsg == WM_INITDIALOG) {
		PROPSHEETPAGE *pPSP = (PROPSHEETPAGE *)lParam;
		pThis = (CUCIEngine*)pPSP->lParam;
		SetWindowLong(hDlg, DWL_USER, pPSP->lParam);
	} else {
		pThis = (CUCIEngine*)GetWindowLong(hDlg, DWL_USER);
	}
	if (pThis) {
		return pThis->OtherOptionsProc(hDlg, nMsg, wParam, lParam);
	} else {
		return FALSE;
	}
}

void CUCIEngine::SetOtherOptionsValues(HWND hDlg)
{
	SetDlgItemInt(hDlg, IDC_EDT_HASH, m_nHash, TRUE);
	SetDlgItemText(hDlg, IDC_EDT_BOOK, m_sBookPath);
	CheckDlgButton(hDlg, IDC_CHK_LOG, g_bLog ? BST_CHECKED : BST_UNCHECKED);
}

BOOL CUCIEngine::ReadOtherOptionsValues(HWND hDlg, long *pnHash, TCHAR *sBuf, DWORD dwLen, BOOL *pbLog)
{
	*pnHash = GetDlgItemInt(hDlg, IDC_EDT_HASH, NULL, TRUE);
	GetDlgItemText(hDlg, IDC_EDT_BOOK, sBuf, dwLen);
	*pbLog = (IsDlgButtonChecked(hDlg, IDC_CHK_LOG) == BST_CHECKED);
	return TRUE;
}

BOOL CUCIEngine::OtherOptionsProc(HWND hDlg, UINT nMsg, WPARAM wParam, LPARAM lParam)
{
	switch (nMsg) {
	case WM_INITDIALOG:
		{
			SendDlgItemMessage(hDlg, IDC_SPIN_HASH, UDM_SETRANGE, 0, MAKELONG(16, 4));

			SetOtherOptionsValues(hDlg);
		}
		break;
	case WM_DESTROY:
		{
			long nHash;
			TCHAR sBook[MAX_PATH];
			BOOL bLog;
			ReadOtherOptionsValues(hDlg, &nHash, sBook, MAX_PATH, &bLog);
			SendOtherOptionsCommands(FALSE, nHash, sBook, bLog);
		}
		break;
	case WM_COMMAND:
		switch (LOWORD(wParam)) {
		case IDC_EDT_BOOK:
			switch (HIWORD(wParam)) {
			case EN_SETFOCUS:
				SipShowIM(SIPF_ON);
				break;
			case EN_KILLFOCUS:
				SipShowIM(SIPF_OFF);
				break;
			}
			break;
		}
	default:
		return FALSE;
	}
	return TRUE;
}

void CUCIEngine::SetThinking(BOOL bThinking)
{
	m_bThinking = bThinking;
}

void GameFinished(const wchar_t *wsResult, const wchar_t *wsComment)
{
	if (g_pEngine) {
		g_pEngine->GameFinished(wsResult, wsComment);
	}
}

void CUCIEngine::GameFinished(const wchar_t *wsResult, const wchar_t *wsComment)
{
	m_bFinished = TRUE;
	m_pCEBoard->GameFinished(wsResult, wsComment);
}

int IsGameFinished()
{
	if (g_pEngine) {
		return g_pEngine->IsGameFinished();
	} else {
		return TRUE;
	}
}

int CUCIEngine::IsGameFinished()
{
	return m_bFinished;
}
